namespace BookWorm {

	/**
	 * Library that holds all books. It loads the DB and keeps the DB on the
	 * HDD in sync with the changes the user does the each Book-Object
	 */
	public class Library {


		/**
		 * See is_ready ()
		 */
		bool all_ready = false;

		/**
		  * True if and only if all books from the DB are loaded (without covers).
		  */
		bool db_ready = false;

		Sqlite.Database db;

		Gee.ArrayList<Book> booklist = new Gee.ArrayList<Book>();
		
		/**
		 * A list of books that were loaded since the last call of get_delta_books
		 */
		Gee.ArrayList<Book> sync_booklist = new Gee.ArrayList<Book>();
		// This mutex should be locked when we access the sync_booklist
		Mutex sync_mutex;

		/**
		 * Hashmap for quick search of books via the ID.
		 */
		Gee.HashMap<int, Book> search_list = new Gee.HashMap<int, Book> ();
		// This mutex should be locked  if you access the search_list
		Mutex search_mutex;

		/**
		 * List of file-paths that represent ebook-files that should be loaded
		 * in this session.
		 */
		GLib.Queue<string> to_add_paths = new GLib.Queue<string> ();


		private async void new_book_func () throws ThreadError {
			ThreadFunc<void*> run = () => {

				string item = null;

				lock(to_add_paths) {
					item = to_add_paths.pop_head ();
				}
				while (item != null) {
					add_book (item);
					lock(to_add_paths) {
						item = to_add_paths.pop_head ();
					}
				}
				lock(all_ready) {
					all_ready = true;
				}
				return null;
			};
			new Thread<void*>("new_book", run);
			yield;
		}
		
		
		/**
		 * Loads the DB and parses all books. Then triggers the loading
		 * of all new books.
		 */
		private async void cache_book_func () throws ThreadError {
			ThreadFunc<void*> run = () => {
				load_db ();
				load_books ();

				new_book_func.begin ();

				//While we parse new books we can load all the covers.
				foreach(Book b in booklist) {
					b.reload_cover ();
				}

				lock(all_ready) {
					all_ready = true;
				}
				return null;
			};
			new Thread<void*>("cache_book", run);
			yield;
		}

		/**
		 * Starts caching the books and returns instantly.
		 * NEVER call this method more then once n a library.
		 */
		public void cache_books_async () {
			//TODO assert that we call this only once...
			cache_book_func.begin ();
		}

		private int load_db () {
			string errmsg;

			string basepath = BookWorm.get_data_dir ();
			DirUtils.create_with_parents (basepath, 0775);
			// Open a database:
			int ec = Sqlite.Database.open (basepath + "library.db", out db);
			if (ec != Sqlite.OK) {
				stderr.printf ("Can't open database: %d: %s\n", db.errcode (), db.errmsg ());
				return -1;
			}

			// Insert test data:
			string query = """
				CREATE TABLE if not exists Books (
					id		INT		PRIMARY KEY		NOT NULL,
					name	TEXT					NOT NULL,
					desc	TEXT					NOT NULL,
					author	TEXT					NOT NULL,
					tags	TEXT					NOT NULL,
					bookgroup	TEXT				NOT NULL,
					language	TEXT				NOT NULL,
					publication_date	TEXT		NOT NULL,
					isbn	TEXT					NOT NULL,
					restore_data	TEXT			NOT NULL,
					hash	TEXT					NOT NULL
				);
				""";
			ec = db.exec (query, null, out errmsg);
			if (ec != Sqlite.OK) {
				stderr.printf ("Error: %s\n", errmsg);
				return -2;
			}
			return 0;
		}

		/**
		  * True if and only if all books (including newly added books)
		  * are added.
		  */
		public bool is_ready () {
			lock (all_ready) {
				return all_ready;
			}
		}

		/**
		 * Returns a Book with the specified ID or null if we don't have such a
		 * book loaded so far.
		 */
		public Book? get_book (int id) {
			Book b;
			search_mutex.lock ();
			b = search_list.get (id);
			search_mutex.unlock ();
			return b;
		}

		private void write_book_to_db (Book b) {
			string errmsg;
			string query =
				@"INSERT INTO Books (id, name, desc, author, tags, bookgroup, language, publication_date," +
				@"isbn, restore_data, hash) VALUES ($(b.id), '$(b.name)', '$(b.desc)', '$(b.author)'," +
				@"'$(b.tags)', '$(b.bookgroup)', '$(b.language)', '$(b.publication_date)'," +
				@"'$(b.isbn)', '$(b.restore_data)', '$(b.hash)');";
			debug(query);
			int ec = db.exec (query, null, out errmsg);
			if (ec != Sqlite.OK) {
				warning ("Error: %s\n", errmsg);
				return;
			}
		}

		private void update_book_in_db (Book b) {
			string errmsg;
			string query =
				@"UPDATE Books "+ 
				@"SET name='$(b.name)', desc='$(b.desc)', author='$(b.author)'," +
				@"tags='$(b.tags)', bookgroup='$(b.bookgroup)', language='$(b.language)', publication_date='$(b.publication_date)'," +
				@"isbn='$(b.isbn)', restore_data='$(b.restore_data)', hash='$(b.hash)'" +
				@"WHERE id=$(b.id);";
			debug(query);
			int ec = db.exec (query, null, out errmsg);
			if (ec != Sqlite.OK) {
				warning ("Error: %s\n", errmsg);
				return;
			}
		}

		/**
		 * Starts loading all books
		 */
		private void load_books () {
			Sqlite.Statement stmt;
			const string prepared_query_str = "SELECT * FROM Books";
			int ec = db.prepare_v2 (prepared_query_str, prepared_query_str.length, out stmt);
			if (ec != Sqlite.OK) {
				stderr.printf ("Error: %d: %s\n", db.errcode (), db.errmsg ());
				//FIXME throw a exception
			}

			while (stmt.step () == Sqlite.ROW) {
				Book b =
					new Book (int.parse (stmt.column_text (0)),
								stmt.column_text (1),
								stmt.column_text (2),
								stmt.column_text (3),
								stmt.column_text (4),
								stmt.column_text (5),
								stmt.column_text (6),
								stmt.column_text (7),
								stmt.column_text (8),
								stmt.column_text (9),
								stmt.column_text (10));
				b.update_db_entry.connect (() => {
					update_book_in_db (b);
				});
				sync_mutex.lock ();
				booklist.add (b);
				sync_mutex.unlock ();
				search_mutex.lock ();
				search_list.set (b.id, b);
				search_mutex.unlock ();
			}

			// Reset the statement to rebind parameters:
			stmt.reset ();
			lock(db_ready) {
				db_ready = true;
			}
		}

		public Gee.ArrayList<Book> get_delta_books () {
			var result = new Gee.ArrayList<Book>();
			sync_mutex.lock ();
			foreach (Book b in booklist) {
				if (!sync_booklist.contains (b)) {
					sync_booklist.add (b);
					result.add (b);
				}
			}
			sync_mutex.unlock ();
			return result;
		}

		/**
		 * Returns a free id as a integer. This function should never fail in
		 * a real scenario. I mean, who the hell has more then 2 billion ebooks?
		 */ 
		public int get_free_id () {
			for(int i = int.MIN; i < int.MAX; i++) {
				if(!search_list.has_key (i)) {
					return i;
				}
			}
			error ("Oh, you are out of indizes!");
		}

		/**
		 * Requests that the Library should load the file somewhen in the future
		 * into the DB.
		 */
		public void request_book (string path) {
			lock(to_add_paths) {
				to_add_paths.push_tail (path);
				lock(db_ready) {
					if(db_ready)
						new_book_func.begin ();;
				}
			}
		}

		/**
		 * Loads a book into the DB. Also moves files such as the cover
		 * into the right place.
		 */
		private void add_book (string path) {
			Book b = new Book.from_epub (get_free_id (), path);

			sync_mutex.lock ();
			foreach(Book check in booklist) {
				if(check.hash == b.hash || check.name == b.name) {
					warning ("Dublicate: " + b.to_string () + " => " +
								check.to_string ());
					sync_mutex.unlock ();
					return;
				}
			}
			booklist.add (b);
			b.update_db_entry.connect (() => {
				update_book_in_db (b);
			});
			
			sync_mutex.unlock ();

			search_mutex.lock ();
			search_list.set (b.id, b);
			search_mutex.unlock();

			File file1 = File.new_for_path (path);
			File file2 = File.new_for_path (BookWorm.get_data_dir()
									+ "ebooks/" + b.id.to_string ());
			try {
				file1.copy (file2, 0, null);
			} catch (Error e) {
				warning ("Errddor: %s\n", e.message);
			}
			write_book_to_db (b);
		}

	}
}
